@@ -1562,18 +1562,18 @@ constant shift greater than word length")))
      (:translate ,function)))
 
 (macrolet ((def (name -c-p)
-             (let ((fun64 (intern (format nil "~S-MOD64" name)))
-                   (vopu (intern (format nil "FAST-~S/UNSIGNED=>UNSIGNED" name)))
-                   (vopcu (intern (format nil "FAST-~S-C/UNSIGNED=>UNSIGNED" name)))
-                   (vopf (intern (format nil "FAST-~S/FIXNUM=>FIXNUM" name)))
-                   (vopcf (intern (format nil "FAST-~S-C/FIXNUM=>FIXNUM" name)))
-                   (vop64u (intern (format nil "FAST-~S-MOD64/WORD=>UNSIGNED" name)))
-                   (vop64f (intern (format nil "FAST-~S-MOD64/FIXNUM=>FIXNUM" name)))
-                   (vop64cu (intern (format nil "FAST-~S-MOD64-C/WORD=>UNSIGNED" name)))
-                   (vop64cf (intern (format nil "FAST-~S-MOD64-C/FIXNUM=>FIXNUM" name)))
-                   (funfx (intern (format nil "~S-MODFX" name)))
-                   (vopfxf (intern (format nil "FAST-~S-MODFX/FIXNUM=>FIXNUM" name)))
-                   (vopfxcf (intern (format nil "FAST-~S-MODFX-C/FIXNUM=>FIXNUM" name))))
+             (let ((fun64   (symbolicate name "-MOD64"))
+                   (funfx   (symbolicate name "-MODFX"))
+                   (vopu    (symbolicate "FAST-" name "/UNSIGNED=>UNSIGNED"))
+                   (vopcu   (symbolicate "FAST-" name "-C/UNSIGNED=>UNSIGNED"))
+                   (vopf    (symbolicate "FAST-" name "/FIXNUM=>FIXNUM"))
+                   (vopcf   (symbolicate "FAST-" name "-C/FIXNUM=>FIXNUM"))
+                   (vop64u  (symbolicate "FAST-" name "-MOD64/WORD=>UNSIGNED"))
+                   (vop64f  (symbolicate "FAST-" name "-MOD64/FIXNUM=>FIXNUM"))
+                   (vop64cu (symbolicate "FAST-" name "-MOD64-C/WORD=>UNSIGNED"))
+                   (vop64cf (symbolicate "FAST-" name "-MOD64-C/FIXNUM=>FIXNUM"))
+                   (vopfxf  (symbolicate "FAST-" name "-MODFX/FIXNUM=>FIXNUM"))
+                   (vopfxcf (symbolicate "FAST-" name "-MODFX-C/FIXNUM=>FIXNUM")))
                (declare (ignore vop64cf)) ; maybe someone will want it some day
                `(progn
                   (define-modular-fun ,fun64 (x y) ,name :untagged nil 64)
@@ -1749,16 +1749,15 @@ constant shift greater than word length")))
     (inst test digit digit)))
 
 
-;;; For add and sub with carry the sc of carry argument is any-reg so
-;;; that it may be passed as a fixnum or word and thus may be 0, 1, or
-;;; 8. This is easy to deal with and may save a fixnum-word
-;;; conversion.
+;;; For add and sub with carry, the sc of carry argument is unsigned-reg
+;;; or any-reg so that it may be passed either as tagged or untagged.
+;;; This is easy to deal with and may save a fixnum-word conversion.
 (define-vop (add-w/carry)
   (:translate sb-bignum:%add-with-carry)
   (:policy :fast-safe)
   (:args (a :scs (unsigned-reg) :target result)
          (b :scs (unsigned-reg unsigned-stack) :to :eval)
-         (c :scs (any-reg) :target temp))
+         (c :scs (any-reg unsigned-reg control-stack) :target temp))
   (:arg-types unsigned-num unsigned-num positive-fixnum)
   (:temporary (:sc any-reg :from (:argument 2) :to :eval) temp)
   (:results (result :scs (unsigned-reg) :from (:argument 0))
@@ -1769,8 +1768,8 @@ constant shift greater than word length")))
     (move temp c)
     (inst neg temp) ; Set the carry flag to 0 if c=0 else to 1
     (inst adc result b)
-    (inst mov carry 0)
-    (inst adc carry carry)))
+    (inst set carry :c)
+    (inst and :dword carry 1)))
 
 ;;; Note: the borrow is 1 for no borrow and 0 for a borrow, the opposite
 ;;; of the x86-64 convention.
@@ -1779,7 +1778,7 @@ constant shift greater than word length")))
   (:policy :fast-safe)
   (:args (a :scs (unsigned-reg) :to :eval :target result)
          (b :scs (unsigned-reg unsigned-stack) :to :result)
-         (c :scs (any-reg control-stack)))
+         (c :scs (any-reg unsigned-reg control-stack)))
   (:arg-types unsigned-num unsigned-num positive-fixnum)
   (:results (result :scs (unsigned-reg) :from :eval)
             (borrow :scs (unsigned-reg)))
@@ -1789,7 +1788,7 @@ constant shift greater than word length")))
     (move result a)
     (inst sbb result b)
     (inst mov borrow 1)
-    (inst sbb borrow 0)))
+    (inst sbb :dword borrow 0)))
 
 
 (define-vop (bignum-mult-and-add-3-arg)
@@ -2102,7 +2101,7 @@ constant shift greater than word length")))
 
 (in-package "SB-C")
 
-(defun *-transformer (y node)
+(defun *-transformer (y node fun)
   (cond
     ((= y (ash 1 (integer-length y)))
      ;; there's a generic transform for y = 2^k
@@ -2110,7 +2109,7 @@ constant shift greater than word length")))
     ((member y '(3 5 9))
      ;; we can do these multiplications directly using LEA
      (delay-ir1-transform node :constraint)
-     `(%lea x x ,(1- y) 0))
+     `(,fun x x ,(1- y) 0))
     (t
      ;; A normal 64-bit multiplication takes 4 cycles on Athlon 64/Opteron.
      ;; Optimizing multiplications (other than the above cases) to
@@ -2128,14 +2127,14 @@ constant shift greater than word length")))
                  :important nil
                  :node node)
   "recode as leas, shifts and adds"
-  (*-transformer (lvar-value y) node))
+  (*-transformer (lvar-value y) node '%lea))
 (deftransform sb-vm::*-mod64
     ((x y) ((unsigned-byte 64) (constant-arg (unsigned-byte 64)))
      (unsigned-byte 64)
      :important nil
      :node node)
   "recode as leas, shifts and adds"
-  (*-transformer (lvar-value y) node))
+  (*-transformer (lvar-value y) node 'sb-vm::%lea-mod64))
 
 (deftransform * ((x y)
                  (fixnum (constant-arg (unsigned-byte 64)))
@@ -2143,11 +2142,11 @@ constant shift greater than word length")))
                  :important nil
                  :node node)
   "recode as leas, shifts and adds"
-  (*-transformer (lvar-value y) node))
+  (*-transformer (lvar-value y) node '%lea))
 (deftransform sb-vm::*-modfx
     ((x y) (fixnum (constant-arg (unsigned-byte 64)))
      fixnum
      :important nil
      :node node)
   "recode as leas, shifts and adds"
-  (*-transformer (lvar-value y) node))
+  (*-transformer (lvar-value y) node 'sb-vm::%lea-modfx))
