@@ -39,7 +39,9 @@
 
 #include "validate.h"
 #include "gc-internal.h"
+#include "gc-private.h"
 #include "pseudo-atomic.h"
+#include "code.h"
 
 #include <errno.h>
 
@@ -83,12 +85,16 @@ open_binary(char *filename, int mode)
 
 #if defined(LISP_FEATURE_LINUX) && defined(LISP_FEATURE_IMMOBILE_CODE)
 #define ELFCORE 1
-extern __attribute__((weak)) lispobj
- __lisp_code_start, lisp_jit_code, __lisp_code_end, __lisp_linkage_values;
-static inline boolean code_in_elf() { return &__lisp_code_start != 0; }
-#else
+#elif !defined(ELFCORE)
 #define ELFCORE 0
-static inline boolean code_in_elf() { return 0; }
+#endif
+
+#if !ELFCORE
+int lisp_code_in_elf() { return 0; }
+#else
+extern __attribute__((weak)) lispobj
+ lisp_code_start, lisp_jit_code, lisp_code_end, lisp_linkage_values;
+int lisp_code_in_elf() { return &lisp_code_start != 0; }
 #endif
 
 /* Search 'filename' for an embedded core.  An SBCL core has, at the
@@ -133,7 +139,7 @@ search_for_embedded_core(char *filename, struct memsize_options *memsize_options
             read(fd, &header, lispobj_size) != lispobj_size || header != CORE_MAGIC)
             core_start = -1; // reset to invalid
     }
-#if ELFCORE
+#if ELFCORE && !defined(LISP_FEATURE_DARWIN)
     // Specifying "--core" as an ELF file with a lisp.core section doesn't work.
     // (There are bunch of reasons) So only search for a core section if this
     // is an implicit search for a core embedded in an executable.
@@ -153,6 +159,7 @@ search_for_embedded_core(char *filename, struct memsize_options *memsize_options
             && optarray[0] == RUNTIME_OPTIONS_MAGIC) {
             memsize_options->dynamic_space_size = optarray[2];
             memsize_options->thread_control_stack_size = optarray[3];
+            memsize_options->thread_tls_bytes = optarray[4];
             memsize_options->present_in_core = 1;
         }
     }
@@ -374,15 +381,17 @@ static void adjust_pointers(lispobj *where, sword_t n_words, struct heap_adjust*
 
 #include "var-io.h"
 #include "unaligned.h"
-static void __attribute__((unused))
-adjust_code_refs(struct heap_adjust* adj, struct code* code, lispobj original_vaddr)
+static void
+adjust_code_refs(struct heap_adjust __attribute__((unused)) *adj,
+                 struct code __attribute__((unused)) *code,
+                 lispobj __attribute__((unused)) original_vaddr)
 {
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
-    // Dynamic space is a always relocated before immobile space,
+    // Dynamic space always gets relocated before immobile space does,
     // and dynamic space code does not use fixups (except on 32-bit x86).
     // So if we're here, it must be to relocate an immobile object.
     // If 'code->fixups' is a bignum, the pointer itself was already fixed up.
-    char* instructions = (char*)((lispobj*)code + code_header_words(code->header));
+    char* instructions = code_text_start(code);
     struct varint_unpacker unpacker;
 
     varint_unpacker_init(&unpacker, code->fixups);
@@ -395,6 +404,8 @@ adjust_code_refs(struct heap_adjust* adj, struct code* code, lispobj original_va
         void* fixup_where = instructions + loc;
         lispobj ptr = UNALIGNED_LOAD32(fixup_where);
         lispobj adjusted = ptr + calc_adjustment(adj, ptr);
+        if (!(adjusted <= UINT32_MAX))
+            lose("Absolute fixup @ %p exceeds 32 bits", fixup_where);
         if (adjusted != ptr)
             FIXUP32(UNALIGNED_STORE32(fixup_where, adjusted), fixup_where);
     }
@@ -406,14 +417,17 @@ adjust_code_refs(struct heap_adjust* adj, struct code* code, lispobj original_va
         loc += prev_loc;
         prev_loc = loc;
         void* fixup_where = instructions + loc;
-        lispobj rel32operand = UNALIGNED_LOAD32(fixup_where);
-        lispobj adjusted = rel32operand - displacement;
+        int32_t rel32operand = UNALIGNED_LOAD32(fixup_where);
+        sword_t adjusted = (sword_t)rel32operand - displacement;
+        if (!(adjusted >= INT32_MIN && adjusted <= INT32_MAX))
+            lose("Relative fixup @ %p exceeds 32 bits", fixup_where);
         FIXUP_rel(UNALIGNED_STORE32(fixup_where, adjusted), fixup_where);
     }
 #endif
 }
 
-static inline void fix_fun_header_layout(lispobj* fun, struct heap_adjust* adj)
+static inline void fix_fun_header_layout(lispobj __attribute__((unused)) *fun,
+                                         struct heap_adjust __attribute__((unused)) *adj)
 {
 #if defined(LISP_FEATURE_COMPACT_INSTANCE_HEADER) && defined(LISP_FEATURE_64_BIT)
     lispobj ptr = function_layout(fun);
@@ -535,18 +549,21 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
 #endif
             continue;
         case CODE_HEADER_WIDETAG:
+            if (filler_obj_p(where)) {
+                if (where[2]) adjust_word_at(where+2, adj);
+                continue;
+            }
             // Fixup the constant pool. The word at where+1 is a fixnum.
-            adjust_pointers(where+2, code_header_words(header_word)-2, adj);
-            // Fixup all embedded simple-funs
             code = (struct code*)where;
+            adjust_pointers(where+2, code_header_words(code)-2, adj);
+            // Fixup all embedded simple-funs
             for_each_simple_fun(i, f, code, 1, {
                 fix_fun_header_layout((lispobj*)f, adj);
 #if FUN_SELF_FIXNUM_TAGGED
-                if (f->self != (lispobj)f->code)
-                    FIXUP(f->self = (lispobj)f->code, &f->self);
-                adjust_pointers(SIMPLE_FUN_SCAV_START(f), SIMPLE_FUN_SCAV_NWORDS(f), adj);
+                if (f->self != (lispobj)f->insts)
+                    FIXUP(f->self = (lispobj)f->insts, &f->self);
 #else
-                adjust_pointers(&f->self, (lispobj*)f->code - &f->self, adj);
+                adjust_pointers(&f->self, 1, adj);
 #endif
             });
             {
@@ -569,19 +586,24 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
             adjust_word_at(where+1, adj);
 #endif
             break;
-        // Vectors require extra care because of EQ-based hashing.
+        // Vectors require extra care because of address-based hashing.
         case SIMPLE_VECTOR_WIDETAG:
-          if (is_vector_subtype(*where, VectorValidHashing)) {
+          if (is_vector_subtype(*where, VectorAddrHashing)) {
               struct vector* v = (struct vector*)where;
-              gc_assert(v->length > 0 &&
-                        !(v->length & make_fixnum(1)) && // length must be even
-                        instancep(v->data[0]));
+              // If you could make a hash-table vector with space for exactly 1 k/v pair,
+              // it would have length 5.
+              gc_assert(v->length >= make_fixnum(5));
               lispobj* data = (lispobj*)v->data;
-              adjust_pointers(&data[0], 1, adj); // adjust the hash-table structure
+              adjust_pointers(&data[fixnum_value(v->length)-1], 1, adj);
+              int hwm = KV_PAIRS_HIGH_WATER_MARK(data);
               boolean needs_rehash = 0;
-              lispobj *where = &data[2], *end = &data[fixnum_value(v->length)];
+              lispobj *where = &data[2], *end = &data[2*(hwm+1)];
               // Adjust the elements, checking for need to rehash.
               for ( ; where < end ; where += 2) {
+                  // Really we should use the hash values to figure out which
+                  // keys were address-sensitive. This simply overapproximates
+                  // by assuming that any change forces rehash.
+                  // (Similar issue exists in 'fixup_space' in immobile-space.c)
                   lispobj ptr = *where; // key
                   if (is_lisp_pointer(ptr) && (delta = calc_adjustment(adj, ptr)) != 0) {
                       FIXUP(*where = ptr + delta, where);
@@ -592,7 +614,7 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
                       FIXUP(where[1] = ptr + delta, where+1);
               }
               if (needs_rehash) // set v->data[1], the need-to-rehash bit
-                  data[1] = make_fixnum(1);
+                  KV_PAIRS_REHASH(data) |= make_fixnum(1);
               continue;
           }
         // All the array header widetags.
@@ -631,6 +653,9 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         case COMPLEX_DOUBLE_FLOAT_WIDETAG:
 #ifdef SIMD_PACK_WIDETAG
         case SIMD_PACK_WIDETAG:
+#endif
+#ifdef SIMD_PACK_256_WIDETAG
+        case SIMD_PACK_256_WIDETAG:
 #endif
             continue;
         default:
@@ -674,12 +699,12 @@ static void relocate_heap(struct heap_adjust* adj)
     // Assembler routines move with varyobj space.
     // (They're in readonly space if no immobile space)
     struct code* code = get_asm_routine_code_component();
-    lispobj* jump_table = (lispobj*)code + code_header_words(code->header);
+    lispobj* jump_table = (lispobj*)code_text_start(code);
     for ( ; *jump_table ; ++jump_table )
         adjust_word_at(jump_table, adj);
     // Pointers within varyobj space to varyobj space do not need adjustment
     // so remove any delta before performing the relocation pass on this space.
-    if (code_in_elf())
+    if (lisp_code_in_elf())
         adj->range[2].delta = 0; // FIXME: isn't this this already the case?
     relocate_space(VARYOBJ_SPACE_START, varyobj_free_pointer, adj);
 #endif
@@ -700,6 +725,12 @@ set_adjustment(struct heap_adjust* adj,
 }
 #endif
 
+#if defined(LISP_FEATURE_ELF) && defined(LISP_FEATURE_IMMOBILE_SPACE)
+    extern int apply_pie_relocs(long,long,int);
+#else
+#   define apply_pie_relocs(dummy1,dummy2,dummy3) (0)
+#endif
+
 static void
 process_directory(int count, struct ndir_entry *entry,
                   int fd, os_vm_offset_t file_offset,
@@ -735,19 +766,21 @@ process_directory(int count, struct ndir_entry *entry,
     };
 
 #if ELFCORE
-    if (&__lisp_code_start) {
-        VARYOBJ_SPACE_START = (uword_t)&__lisp_code_start;
+    if (&lisp_code_start) {
+        VARYOBJ_SPACE_START = (uword_t)&lisp_code_start;
         varyobj_free_pointer = &lisp_jit_code;
-        varyobj_space_size = (uword_t)&__lisp_code_end - VARYOBJ_SPACE_START;
+        varyobj_space_size = (uword_t)&lisp_code_end - VARYOBJ_SPACE_START;
         spaces[IMMOBILE_VARYOBJ_CORE_SPACE_ID].len = varyobj_space_size;
-        gc_assert(varyobj_free_pointer >= (lispobj*)VARYOBJ_SPACE_START);
-#if !ENABLE_PAGE_PROTECTION
+        if (varyobj_free_pointer < (lispobj*)VARYOBJ_SPACE_START
+            || !PTR_IS_ALIGNED(&lisp_code_end, 4096))
+            lose("ELF core alignment bug. Check for proper padding in 'editcore'");
+#ifdef DEBUG_COREPARSE
         printf("Lisp code present in executable @ %lx:%lx (freeptr=%p)\n",
-               (uword_t)&__lisp_code_start, (uword_t)&__lisp_code_end,
+               (uword_t)&lisp_code_start, (uword_t)&lisp_code_end,
                varyobj_free_pointer);
 #endif
-        // Prefill the Lisp linkage table so that we can (pending some additional work)
-        // remove "-ldl" from the linker options when making a shrinkwrapped executable.
+        // Prefill the Lisp linkage table so that shrinkwrapped executables which link in
+        // all their C library dependencies can avoid linking with -ldl.
         // All data references are potentially needed because aliencomp doesn't emit
         // SAP-REF-n in a way that admits elision of the linkage entry. e.g.
         //     MOV RAX, [#x20200AA0] ; some_c_symbol
@@ -757,7 +790,7 @@ process_directory(int count, struct ndir_entry *entry,
         // but that's more of a change to the asm instructions than I'm comfortable making;
         // whereas "CALL linkage_entry_for_f" -> "CALL f" is quite straightforward.
         // (Rarely would a jmp indirection be used; maybe for newly compiled code?)
-        lispobj* ptr = &__lisp_linkage_values;
+        lispobj* ptr = &lisp_linkage_values;
         gc_assert(ptr);
         char *link_target = (char*)(intptr_t)LINKAGE_TABLE_SPACE_START;
         int count;
@@ -783,6 +816,11 @@ process_directory(int count, struct ndir_entry *entry,
     for ( ; --count>= 0; ++entry) {
         long id = entry->identifier;
         uword_t addr = entry->address;
+#ifdef DEBUG_COREPARSE
+        printf("space %d @ %10lx pg=%4d+%4d nwords=%ld\n",
+               (int)id, addr, (int)entry->data_page, (int)entry->page_count,
+               entry->nwords);
+#endif
         int compressed = id & DEFLATED_CORE_SPACE_ID_FLAG;
         id -= compressed;
         if (id < 1 || id > MAX_CORE_SPACE_ID)
@@ -922,10 +960,6 @@ process_directory(int count, struct ndir_entry *entry,
             set_alloc_pointer((lispobj)free_pointer);
 
             anon_dynamic_space_start = (os_vm_address_t)(addr + len);
-            /* This assertion safeguards the test in zero_pages_with_mmap()
-             * which trusts that if addr > anon_dynamic_space_start
-             * then addr did not come from any file mapping. */
-            gc_assert((lispobj)anon_dynamic_space_start > STATIC_SPACE_END);
         }
     }
 
@@ -938,9 +972,13 @@ process_directory(int count, struct ndir_entry *entry,
     set_adjustment(adj, FIXEDOBJ_SPACE_START, // actual
                    spaces[IMMOBILE_FIXEDOBJ_CORE_SPACE_ID].base, // expected
                    spaces[IMMOBILE_FIXEDOBJ_CORE_SPACE_ID].len);
-    set_adjustment(adj, VARYOBJ_SPACE_START, // actual
-                   spaces[IMMOBILE_VARYOBJ_CORE_SPACE_ID].base, // expected
-                   spaces[IMMOBILE_VARYOBJ_CORE_SPACE_ID].len);
+    if (!apply_pie_relocs(VARYOBJ_SPACE_START
+                          - spaces[IMMOBILE_VARYOBJ_CORE_SPACE_ID].base,
+                          DYNAMIC_SPACE_START - spaces[DYNAMIC_CORE_SPACE_ID].base,
+                          fd))
+        set_adjustment(adj, VARYOBJ_SPACE_START, // actual
+                       spaces[IMMOBILE_VARYOBJ_CORE_SPACE_ID].base, // expected
+                       spaces[IMMOBILE_VARYOBJ_CORE_SPACE_ID].len);
 #    endif
 #  else
     set_adjustment(adj, DYNAMIC_0_SPACE_START, // actual
@@ -948,7 +986,7 @@ process_directory(int count, struct ndir_entry *entry,
                    spaces[DYNAMIC_CORE_SPACE_ID].len);
 #  endif // LISP_FEATURE_GENCGC
     if (adj->range[0].delta | adj->range[1].delta | adj->range[2].delta) {
-        if (adj->range[1].delta && code_in_elf())
+        if (adj->range[1].delta && lisp_code_in_elf())
             lose("Relocation of fixedobj space not supported with ELF core.\n\
 Please report this as a bug");
         relocate_heap(adj);
@@ -998,6 +1036,8 @@ extern void gc_load_corefile_ptes(core_entry_elt_t, core_entry_elt_t,
 #define gc_load_corefile_ptes(dummy1,dummy2,dummy3,dummy4)
 #endif
 
+static void sanity_check_loaded_core(lispobj);
+
 /* 'merge_core_pages': Tri-state flag to determine whether we attempt to mark
  * pages as targets for virtual memory deduplication via MADV_MERGEABLE.
  * 1: Yes
@@ -1065,6 +1105,13 @@ load_core_file(char *file, os_vm_offset_t file_offset, int merge_core_pages)
         case END_CORE_ENTRY_TYPE_CODE:
             free(header);
             close(fd);
+#ifdef LISP_FEATURE_SB_THREAD
+            if ((int)SymbolValue(FREE_TLS_INDEX,0) >= dynamic_values_bytes) {
+                dynamic_values_bytes = (int)SymbolValue(FREE_TLS_INDEX,0) * 2;
+                // fprintf(stderr, "NOTE: TLS size increased to %x\n", dynamic_values_bytes);
+            }
+#endif
+            sanity_check_loaded_core(initial_function);
             return initial_function;
         case RUNTIME_OPTIONS_MAGIC: break; // already processed
         default:
@@ -1082,15 +1129,14 @@ os_vm_address_t get_asm_routine_by_name(const char* name)
     lispobj ht = CONS(code->debug_info)->car;
     if (ht) {
         struct vector* table =
-            VECTOR(((struct hash_table*)native_pointer(ht))->table);
+            VECTOR(((struct hash_table*)native_pointer(ht))->pairs);
         lispobj sym;
         int i;
         for (i=2 ; i < fixnum_value(table->length) ; i += 2)
             if (lowtag_of(sym = table->data[i]) == OTHER_POINTER_LOWTAG
                 && widetag_of(&SYMBOL(sym)->header) == SYMBOL_WIDETAG
                 && !strcmp(name, (char*)(VECTOR(SYMBOL(sym)->name)->data)))
-                return code_header_words(code->header)*N_WORD_BYTES
-                    + fixnum_value(CONS(table->data[i+1])->car) + (os_vm_address_t)code;
+                return code_text_start(code) + fixnum_value(CONS(table->data[i+1])->car);
         // Something is wrong if we have a hashtable but find nothing.
         fprintf(stderr, "WARNING: get_asm_routine_by_name(%s) failed\n",
                 name);
@@ -1104,3 +1150,218 @@ void asm_routine_poke(const char* routine, int offset, char byte)
     if (address)
         address[offset] = byte;
 }
+
+// Caution: use at your own risk
+#undef DEBUG_CORE_LOADING
+#ifdef DEBUG_CORE_LOADING
+#ifdef LISP_FEATURE_CHENEYGC
+#  error "Can't define DEBUG_CORE_LOADING for cheneygc"
+#endif
+#include "hopscotch.h"
+#include "genesis/cons.h"
+#include "genesis/layout.h"
+#include "genesis/gc-tables.h"
+#include "code.h"
+#include "gc-private.h"
+
+struct visitor {
+    // one item per value of widetag>>2
+    // element 0 is for conses, element 64 is for totals.
+    struct {
+        int count;
+        int words;
+    } headers[65], sv_subtypes[3];
+    struct hopscotch_table *reached;
+};
+
+#define RECURSE(x) if(is_lisp_pointer(x))graph_visit(ptr,x,seen)
+static void graph_visit(lispobj __attribute__((unused)) referer,
+                        lispobj ptr,
+                        struct hopscotch_table* seen)
+{
+    if (lowtag_of(ptr) == FUN_POINTER_LOWTAG
+        && widetag_of(FUNCTION(ptr)) == SIMPLE_FUN_WIDETAG)
+        ptr = fun_code_tagged(FUNCTION(ptr));
+    if (hopscotch_get(seen, ptr, 0))
+        return;
+    hopscotch_insert(seen, ptr, 1);
+    lispobj layout, bitmap, *obj;
+    int nwords, i;
+    if (lowtag_of(ptr) == LIST_POINTER_LOWTAG) {
+        RECURSE(CONS(ptr)->car);
+        RECURSE(CONS(ptr)->cdr);
+    } else switch (widetag_of(obj = native_pointer(ptr))) {
+        case SIMPLE_VECTOR_WIDETAG:
+            nwords = fixnum_value(obj[1]); // vector length
+            for(i=0; i<nwords; ++i) RECURSE(obj[i+2]);
+            break;
+        // In all the following cases except for CODE, 'nwords' is the count
+        // of payload words (following the header), so we iterate up to and
+        // including that word index. For example, if there are 2 payload words,
+        // then we scan word indices 1 and 2 off the object base address.
+        case INSTANCE_WIDETAG:
+            layout = instance_layout(obj);
+            graph_visit(ptr, layout, seen);
+            nwords = instance_length(*obj);
+            bitmap = ((struct layout*)native_pointer(layout))->bitmap;
+            for(i=1; i<=nwords; ++i)
+                if (layout_bitmap_logbitp(i-1, bitmap)) RECURSE(obj[i]);
+            break;
+        case FUNCALLABLE_INSTANCE_WIDETAG:
+            layout = funinstance_layout(obj);
+            graph_visit(ptr, layout, seen);
+            bitmap = ((struct layout*)native_pointer(layout))->bitmap;
+            nwords = SHORT_BOXED_NWORDS(*obj);
+            // We don't need to scan the word at index 1 (the trampoline pointer)
+            // because it either points to the FIN itself or to readonly space.
+            for(i=2; i<=nwords; ++i)
+                if (layout_bitmap_logbitp(i-1, bitmap)) RECURSE(obj[i]);
+            break;
+        case CLOSURE_WIDETAG:
+            // We must scan the closure's trampoline word.
+            // Closures can utilize one payload word beyond what the header
+            // indicates. This is quite sucky and I don't know why I did that.
+            // However, it is correctly accounted for by SHORT_BOXED_NWORDS
+            // which gives you the right number of words to scan.
+            graph_visit(ptr, obj[1] - FUN_RAW_ADDR_OFFSET, seen);
+            nwords = SHORT_BOXED_NWORDS(*obj);
+            for(i=2; i<=nwords; ++i) RECURSE(obj[i]);
+            break;
+        case SYMBOL_WIDETAG:
+        case WEAK_POINTER_WIDETAG:
+            nwords = TINY_BOXED_NWORDS(*obj);
+            for(i=1; i<=nwords; ++i) RECURSE(obj[i]);
+            break;
+        case FDEFN_WIDETAG:
+            RECURSE(obj[1]);
+            RECURSE(obj[2]);
+            RECURSE(fdefn_callee_lispobj((struct fdefn*)obj));
+            break;
+        case CODE_HEADER_WIDETAG:
+            nwords = code_header_words((struct code*)obj);
+            for(i=2; i<nwords; ++i) RECURSE(obj[i]);
+            break;
+        default:
+            if (!leaf_obj_widetag_p(widetag_of(obj))) {
+                nwords = BOXED_NWORDS(*obj);
+                for(i=1; i<=nwords; ++i) RECURSE(obj[i]);
+            }
+      }
+}
+
+static void tally(lispobj ptr, struct visitor* v)
+{
+    if (lowtag_of(ptr) == LIST_POINTER_LOWTAG)
+        ++v->headers[0].count;
+    else {
+        lispobj* obj = native_pointer(ptr);
+        int widetag = widetag_of(obj);
+        int header_index = widetag>>2;
+        int words = OBJECT_SIZE(*obj, obj);
+        ++v->headers[header_index].count;
+        v->headers[header_index].words += words;
+        if (widetag == SIMPLE_VECTOR_WIDETAG) {
+            int st = 0;
+            switch ((*obj >> N_WIDETAG_BITS) & 7) {
+            case subtype_VectorWeak: st = 1; break;
+            case subtype_VectorValidHashing: st = 2; break;
+            }
+            ++v->sv_subtypes[st].count;
+            v->sv_subtypes[st].words += words;
+        }
+    }
+}
+
+static uword_t visit(lispobj* where, lispobj* limit, uword_t arg)
+{
+    struct visitor* v = (struct visitor*)arg;
+    lispobj* obj = where;
+    while (obj < limit) {
+        lispobj ptr = compute_lispobj(obj);
+        tally(ptr, v);
+        // Weird stuff happens in genesis where this logic thinks that the
+        // cold core holds a ton of unreachable objects. I have no idea why.
+        if (!hopscotch_get(v->reached, ptr, 0))
+            printf("object not reached: %p\n", (void*)ptr);
+        obj += OBJECT_SIZE(*obj, obj);
+    }
+    return 0;
+}
+
+#ifdef LISP_FEATURE_GENCGC
+#define dynamic_space_pointer_p(ptr) (find_page_index((void*)ptr) >= 0)
+#endif
+
+static void sanity_check_loaded_core(lispobj initial_function)
+{
+    struct visitor v[2];
+    struct hopscotch_table reached;
+    memset(v, 0, sizeof v);
+    // Pass 1: Count objects reachable from known roots.
+    hopscotch_create(&reached, HOPSCOTCH_HASH_FUN_DEFAULT,
+                     0, // no values
+                     1<<18, /* initial size */
+                     0);
+    {
+      lispobj* where = (lispobj*)STATIC_SPACE_START;
+      lispobj* end = static_space_free_pointer;
+      while (where<end) {
+        // This falsely treats NIL as 4 conses but it doesn't really matter.
+        // The garbage collectors do too.
+        graph_visit(0, compute_lispobj(where), &reached);
+        where += OBJECT_SIZE(*where, where);
+      }
+    }
+    graph_visit(0, initial_function, &reached); // not otherwise reachable
+    // having computed the reaching graph, tally up the dynamic space objects
+    int key_index;
+    lispobj ptr;
+    for_each_hopscotch_key(key_index, ptr, reached)
+        if (dynamic_space_pointer_p(ptr))
+            tally(ptr, &v[0]);
+    // Pass 2: Count all heap objects
+    v[1].reached = &reached;
+    walk_generation(visit, -1, (uword_t)&v[1]);
+    // Pass 3: Compare
+    // Start with the conses
+    v[0].headers[0].words = v[0].headers[0].count * 2;
+    v[1].headers[0].words = v[1].headers[0].count * 2;
+    printf("-----------------------------------------------|\n");
+    printf("       Graph walk     |         Actual         |\n");
+    printf("----------------------+------------------------|\n");
+    int i;
+    for(i=0; i<=64; ++i) {
+        // print all valid widetags (not unknown) that aren't for immediates,
+        // but always print if nonzero.
+        if (v[1].headers[i].count ||
+            ((strncmp(widetag_names[i], "unk", 3)
+              && (i != CHARACTER_WIDETAG>>2)
+              && (i != SIMPLE_FUN_WIDETAG>>2)
+              && (i != NO_TLS_VALUE_MARKER_WIDETAG>>2)
+              && (i != UNBOUND_MARKER_WIDETAG>>2)))) {
+            int mismatch = v[0].headers[i].count != v[1].headers[i].count;
+            printf("%8d %11d  | %8d %11d   | %s%s\n",
+                   v[0].headers[i].count, v[0].headers[i].words,
+                   v[1].headers[i].count, v[1].headers[i].words,
+                   i<64 ? (i ? widetag_names[i] : "cons") : "TOTAL",
+                   mismatch ? " <<<<" : "");
+            if (i == SIMPLE_VECTOR_WIDETAG>>2) {
+                int j;
+                for(j=1; j <= 2; ++j)
+                    printf("%8d %11d  | %8d %11d   |   %s\n",
+                           v[0].sv_subtypes[j].count, v[0].sv_subtypes[j].words,
+                           v[1].sv_subtypes[j].count, v[1].sv_subtypes[j].words,
+                           j==1 ? "weak" : "hashing");
+
+            }
+            v[0].headers[64].count += v[0].headers[i].count;
+            v[1].headers[64].count += v[1].headers[i].count;
+            v[0].headers[64].words += v[0].headers[i].words;
+            v[1].headers[64].words += v[1].headers[i].words;
+        }
+    }
+    hopscotch_destroy(&reached);
+}
+#else
+static void sanity_check_loaded_core(lispobj __attribute__((unused)) initial_function) {}
+#endif
